UNIT GenRand;

(* This unit contains a number of procedures to generate random variables *)

INTERFACE

PROCEDURE GenNormal(mu, sigma: double; VAR a,b: double);

PROCEDURE GenNormGTZ(mu, sigma: double; VAR a,b: double);

PROCEDURE GenNormKnot(mu, sigma, l, u: double; VAR a,b: double);

PROCEDURE GenLogNormal(mu, sigma: double; VAR a,b: double);

PROCEDURE GenGumbel(dzeta, theta: double; VAR a: double);

PROCEDURE GenExponential(lambda: double; VAR a: double);

PROCEDURE GenGamma(alpha,beta: double; VAR a: double);

PROCEDURE GenPareto(alpha, kappa: double; VAR a: double);

PROCEDURE GenTriang(a,b,c: double; VAR x: double);

IMPLEMENTATION

FUNCTION Log(x: double): double;
(* returns ln(x) *)

BEGIN
     IF (x > 1.0E-38) AND (x < 1.0E38) THEN
        log:= ln(x)
     ELSE IF x < 1.0E-38 THEN
          log:= ln(1.0E-38)
     ELSE
         log:= ln(1.0E38)
END;

FUNCTION Expo(x: double): double;
(* returns exp(x) *)

BEGIN
     IF (x < 88) AND (x > -88) THEN
        expo:= exp(x)
     ELSE IF x < -88 THEN
          expo:= exp(-88)
     ELSE
         expo:= exp(88)
END;

FUNCTION Pow(x,y: double): double;
(* returns x^y *)

BEGIN
     pow:= expo(y*log(x))
END;

PROCEDURE GenNormal(mu, sigma: double; VAR a,b: double);

VAR u,v,w: double;

BEGIN
     u:= 2*random-1;
     v:= 2*random-1;
     w:= sqr(u)+sqr(v);
     IF w>1 THEN
        GenNormal(mu,sigma,a,b)
     ELSE BEGIN
          w:= sqrt(-2*log(w)/w);
          a:= mu + sigma*w*u;
          b:= mu + sigma*w*v
     END
END;

PROCEDURE GenNormGTZ(mu, sigma: double; VAR a,b: double);

BEGIN
     GenNormal(mu,sigma,a,b);
     WHILE (a < 0) OR (b < 0) DO
           GenNormal(mu,sigma,a,b)
END;

PROCEDURE GenNormKnot(mu, sigma, l, u: double; VAR a,b: double);

BEGIN
     GenNormal(mu,sigma,a,b);
     WHILE (a < l) OR (b < l) OR (a > u) OR (b > u) DO
           GenNormal(mu,sigma,a,b)
END;

PROCEDURE GenLogNormal(mu, sigma: double; VAR a,b: double);

BEGIN
     GenNormal(mu,sigma,a,b);
     a:= expo(a);
     b:= expo(b)
END;

PROCEDURE GenGumbel(dzeta, theta: double; VAR a: double);

BEGIN
     a:= dzeta - theta*log(-log(random))
END;

PROCEDURE GenExponential(lambda: double; VAR a: double);

BEGIN
     a:= -lambda*log(random)
END;

PROCEDURE GamLTOne(alpha: double; VAR a: double);

VAR e,b,p,y: double;

BEGIN
     e:= expo(1);
     b:= (e + alpha)/e;
     P:= b*random;
     IF P < 1 THEN BEGIN
        Y:= expo(log(P)/alpha);
        IF expo(-Y) >= random THEN
           a:= Y
        ELSE
            GamLTOne(alpha,a)
     END
     ELSE BEGIN
         Y:= -log((b-P)/alpha);
         IF expo((alpha-1)*log(Y)) >= random THEN
            a:= Y
         ELSE
             GamLTOne(alpha,a)
     END
END;

PROCEDURE GamGTOne(alpha: double; VAR a: double);

VAR ha, b, q, th, d, u1, u2, v, y, z, w: double;

BEGIN
     ha:= 1/sqrt(2*alpha-1);
     b:= alpha - log(4);
     q:= alpha + 1/ha;
     th:= 4.5;
     d:= 1 + log(th);
     u1:= random;
     u2:= random;
     v:= ha*log(u1/(1-u1));
     y:= alpha*expo(v);
     z:= u1*u1*u2;
     w:= b + q*v - y;
     IF (w + d - th*z) >= 0 THEN
        a:= y
     ELSE IF w >= log(z) THEN
          a:= y
     ELSE
         GamGTOne(alpha,a)
END;

PROCEDURE GenGamma(alpha,beta: double; VAR a: double);

BEGIN
     IF alpha < 1 THEN BEGIN
        GamLTOne(alpha,a);
        a:= beta*a
     END
     ELSE IF alpha = 1 THEN BEGIN
          GenExponential(1,a);
          a:= beta*a
     END
     ELSE BEGIN
         GamGTOne(alpha,a);
         a:= beta*a
     END
END;

PROCEDURE GenPareto(alpha, kappa: double; VAR a: double);

BEGIN
     a:= kappa*expo(-log(1-random)/alpha)
END;

PROCEDURE GenTriang(a,b,c: double; VAR x: double);

VAR d,u,h: double;

BEGIN
     d:= (c-a)/(b-a);
     u:= random;
     IF u <= d THEN
        h:= sqrt(d*u)
     ELSE
         h:= 1-sqrt((1-d)*(1-u));
     x:= a+(b-a)*h
END;

END.